const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;
const Cu = Components.utils;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");


// --------------------------------------------------------------------------------------------------
function AAData(domain, type, description, alternative) {
  this.domain = domain.toLowerCase();
  this.type = type;
  this.description = description;
  this.alternative = Number(alternative);
}
// --------------------------------------------------------------------------------------------------
function AADataCompare(data1, data2) {
  var len1 = data1.domain.split(/\./).length;
  var len2 = data2.domain.split(/\./).length;
  if (len1 < len2)
    return 1;
  if (len1 == len2)
    return 0;
  return -1;
}
// --------------------------------------------------------------------------------------------------
const AAEngine = {
  classDescription: 'computerbildabzockschutz engine component',
  classID: Components.ID('{7829cf8e-d39f-415d-a8c7-94f2b8a13356}'),
  contractID: '@computerbildabzockschutz/engine;1',
  _xpcom_factory: {
    createInstance: function(outer, iid) {
      if (outer)
        throw Cr.NS_ERROR_NO_AGGREGATION;
      if (!AAEngine.initialized)
        throw Cr.NS_ERROR_NOT_INITIALIZED;
      return AAEngine;
    }
  },
  QueryInterface: function(iid) {
    if (iid.equals(Ci.nsISupports))
      return AAEngine;
    throw Cr.NS_ERROR_NO_INTERFACE;
  },
  // --------------------------------------------------------------------------------------------------
  ID: null,
  strings: null,
  aadata: new Array(),
  magics: new Array(),
  magicurls: new Array(),
  magicidx: new Array(),
  UPDATE_DIST: 1000 * 60 * 60 * 8, // update distance in ms (8h)
  UPDATE_DIST_CRITICAL: 1000 * 60 * 60 * 24 * 5, // critical update distance in ms (5d)
  //version: 1, // print version
  version: 2, // online version
  // the domains overriden by the user (not persistent)
  timeStamp: null,
  okdomains: new Array(),
  // --------------------------------------------------------------------------------------------------
  updateURL: 'https://addons.computerbild.de/antiabzocke/data/update.php', // sperrliste aktualisieren
  // updateURL: 'http://cbabzocke.j3s.de/ips.xml',
  init : function(ID) {
    this.ID = ID;
    this.strings = Components.classes['@mozilla.org/intl/stringbundle;1']
           .getService(Components.interfaces.nsIStringBundleService)
           .createBundle('chrome://computerbildabzockschutz/locale/computerbildabzockschutz.properties'),
    this.timeStamp = this.getTimeStamp() - 2 * this.UPDATE_DIST_CRITICAL;
    this.loadDomains();
    var timeStamp = this.getTimeStamp();
    if (this.aadata.length == 0 || ((timeStamp - this.timeStamp) > this.UPDATE_DIST)) {
      this.update();
    } else {
      this.log('Domains are up to date');
    }
  },
  // --------------------------------------------------------------------------------------------------
  getData: function(magic) {
    var idx = null;
    for (var i = 0; i < this.magics.length; i++) {
      if (this.magics[i] == magic) {
        var res = this.aadata[this.magicidx[i]];
        res.url = this.magicurls[i];
        return res;
      }
    }
    return null;
  },
  // --------------------------------------------------------------------------------------------------
  getType: function(idx) {
    return this.aadata[idx].type;
  },
  // --------------------------------------------------------------------------------------------------
  getMagic: function(url, idx) {
    for (var i = 0; i < this.magicurls.length; i++) {
      if (this.magicurls[i] == url) {
        return this.magics[i];
      }
    }
    var magic = Math.round(Math.random() * 100000000);
    this.magics[this.magics.length] = magic;
    this.magicurls[this.magicurls.length] = url;
    this.magicidx[this.magicidx.length] = idx;
    return magic;
  },
  // --------------------------------------------------------------------------------------------------
  extractDomain : function(url) {
    var idx = url.indexOf('://');
    if (idx != -1) {
      url = url.substr(idx + 3);
    }
    idx = url.indexOf('/');
    if (idx != -1) {
      url = url.substring(0, idx);
    }
    return url;
  },
  // --------------------------------------------------------------------------------------------------
  update : function(feedback) {
    if (feedback == null) {
      feedback = false;
    }
    var req = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"]
                 .createInstance(Components.interfaces.nsIXMLHttpRequest);
    var url = this.updateURL + '?v=' + this.version;
    req.open('GET', url, true);
    req.onerror = function(e) {
      // this.alert(AAEngine.strings.getFormattedString('updateError',
      // [1, e.target.status]));
    };
    req.onreadystatechange = function(aEvt) {
      if (req.readyState == 4) {
        if (req.responseXML == null
            || (url.toLowerCase().indexOf('http') == 0 && req.status != 200)) {
          var timeStamp = AAEngine.getTimeStamp();
          if ((timeStamp - AAEngine.timeStamp) > AAEngine.UPDATE_DIST_CRITICAL) {
            this.alert(AAEngine.strings.GetStringFromName('oldData'));
          } else if (feedback) {
            this.alert(AAEngine.strings.getFormattedString(
                'updateError', [ 2, req.status ]));
          }
        } else {
          AAEngine.handleUpdate(req.responseXML, feedback);
        }
      }
    };
    req.overrideMimeType('text/xml');
    try {
      req.send(null);
    } catch (e) {
      this.alert(this.strings.formatStringFromName('updateError', [ 3, e ], 2));
    }
  },
  // --------------------------------------------------------------------------------------------------
  handleUpdate : function(xml, feedback) {
    var version = 1;
    var domains = new Array();
    var types = new Array();
    var descriptions = new Array();
    var alternatives = new Array();
    try {
      var list = xml.getElementsByTagName('list')[0];
      version = list.getAttribute('version');
      if (version == null || version == '') {
        throw this.strings.GetStringFromName('missingVersion');
      }
      if (this.version - version > 0) {
        throw this.strings.formatStringFromName('versionTooSmall', [ version,
            this.version ], 2);
      }

      if (version - this.version == 0) {
        this.log('No newer version found');
        if (feedback) {
          this.alert(this.strings.GetStringFromName('dataOk'));
        }
        return;
      }
      var entries = list.getElementsByTagName('entry');
      for ( var i = 0; i < entries.length; i++) {
        var type = '' + entries[i].getAttribute('state');
        if (type != 'blacklist' && type != 'review' && type != 'whitelist') {
          throw this.strings.formatStringFromName('unknownType', [ type ], 1);
        }
        types[types.length] = type;
        var domain = entries[i].getElementsByTagName('url')[0].firstChild.data;
        domain = domain.replace(/^\s*/, '');
        domain = domain.replace(/\s*$/, '');
        domains[domains.length] = domain;

        var description = '';
        if (entries[i].getElementsByTagName('description')[0].firstChild != null) {
          description = entries[i].getElementsByTagName('description')[0].firstChild.data;
          description = description.replace(/^\s*/, '');
          description = description.replace(/\s*$/, '');
          description = description.replace(/\n/g, ' ');
          description = description.replace(/&/g, '&amp;');
          description = description.replace(/</g, '&lt;');
          description = description.replace(/>/g, '&gt;');
          description = description.replace(/"/g, '&quot;');
          description = description.replace(/'/g, '&#39;');
        }
        descriptions[descriptions.length] = description;

        var alternative = '0';
        if (entries[i].getElementsByTagName('alternatives')[0].firstChild != null) {
          alternative = entries[i].getElementsByTagName('alternatives')[0].firstChild.data;
          alternative = alternative.replace(/^\s*/, '');
          alternative = alternative.replace(/\s*$/, '');
        }
        alternatives[alternatives.length] = alternative;
      }
    } catch (e) {
      this.alert(this.strings.formatStringFromName('updateError', [ 4, e ], 2));
      return;
    }

    try {
      for ( var i = 0; i < domains.length; i++) {
        this.handleUpdateEntry(types[i], domains[i], descriptions[i],
            alternatives[i]);
      }
    } catch (e) {
      this.alert(this.strings.formatStringFromName('updateError', [ 5, e ], 2));
    }
    this.version = version;
    this.aadata.sort(AADataCompare);
    this.saveDomains();
    if (feedback) {
      this.alert(this.strings.GetStringFromName('dataRefreshed'));
    }
  },
  // --------------------------------------------------------------------------------------------------
  alert: function(msg) {
    var prompts = Components.classes['@mozilla.org/embedcomp/prompt-service;1']
                                     .getService(Components.interfaces.nsIPromptService);
    prompts.alert(null, 'COMPUTERBILD-Abzockschutz', msg);
  },
  // --------------------------------------------------------------------------------------------------
  handleUpdateEntry : function(type, domain, description, alternative) {
    domain = domain.toLowerCase();
    for ( var i = 0; i < this.aadata.length; i++) {
      if (this.aadata[i].domain == domain) {
        this.aadata[i].description = description;
        this.aadata[i].alternative = alternative;
        this.aadata[i].type = type;
        return;
      }
    }
    this.aadata[this.aadata.length] = new AAData(domain, type, description,
        alternative);
    return;
  },
  // --------------------------------------------------------------------------------------------------
  log: function(msg) {
    try {
      var consoleService = Components.classes['@mozilla.org/consoleservice;1']
          .getService(Components.interfaces.nsIConsoleService);
      consoleService.logStringMessage('COMPUTERBILD-Abzockschutz: ' + msg);
    } catch(e) { }
  },
  // --------------------------------------------------------------------------------------------------
  getTimeStamp : function() {
    var date = new Date();
    return date.getTime();
  },
  // --------------------------------------------------------------------------------------------------
  dump: function(msg) {
    dump(msg);
  },
  // --------------------------------------------------------------------------------------------------
  getProfilePath : function() {
    var dirService = Components.classes['@mozilla.org/file/directory_service;1']
        .getService(Components.interfaces.nsIProperties);
    return dirService.get('ProfD', Components.interfaces.nsIFile);
  },
  // --------------------------------------------------------------------------------------------------
  saveDomains : function() {
    var doc = Components.classes['@mozilla.org/xml/xml-document;1']
                    .createInstance(Components.interfaces.nsIDOMXMLDocument);
    var domains = doc.createElement('ds');
    domains.setAttribute('v', this.encode64(this.version));
    domains.setAttribute('t', this.encode64('' + this.getTimeStamp()));
    doc.appendChild(domains);
    for ( var i = 0; i < this.aadata.length; i++) {
      var domain = doc.createElement('d');
      domain.setAttribute('t', this.encode64(this.aadata[i].type));
      domain.setAttribute('n', this.encode64(this.aadata[i].domain));
      domain.setAttribute('d', this.encode64(this.aadata[i].description));
      domain.setAttribute('a', this.encode64(this.aadata[i].alternative));
      domains.appendChild(domain);
    }
    //var serializer = new XMLSerializer();
    var serializer = Components.classes['@mozilla.org/xmlextras/xmlserializer;1']
                  .createInstance(Components.interfaces.nsIDOMSerializer);
    var foStream = Components.classes['@mozilla.org/network/file-output-stream;1']
        .createInstance(Components.interfaces.nsIFileOutputStream);
    var file = this.getProfilePath();
    file.append('computerbildabzockschutz.xml');
    foStream.init(file, 0x02 | 0x08 | 0x20, 0664, 0); // write, create, truncate
    serializer.serializeToStream(doc, foStream, ''); // rememeber, doc is the
                                                      // DOM tree
    foStream.close();
  },
  // --------------------------------------------------------------------------------------------------
  loadDomains : function() {
    var file = this.getProfilePath();
    file.append('computerbildabzockschutz.xml');
    if (!file.exists()) {
      return;
    }
    try {
      var url = 'file://' + file.path;
      var req = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"]
                .createInstance(Components.interfaces.nsIXMLHttpRequest);
      req.open('GET', url, false);
      req.send(null);
      var doc = req.responseXML;
      var domains = doc.getElementsByTagName('ds')[0];
      var version = this.decode64(domains.getAttribute('v'));
      var timeStamp = this.decode64(domains.getAttribute('t'));
      var domain = domains.firstChild;
      var count = 0;
      while (domain != null) {
        var d = this.decode64(domain.getAttribute('n'));
        var type = this.decode64(domain.getAttribute('t'));
        var description = this.decode64(domain.getAttribute('d'));
        var alternative = this.decode64(domain.getAttribute('a'));
        this.aadata[this.aadata.length] = new AAData(d, type, description,
            alternative);
        ++count;
        domain = domain.nextSibling;
      }
      this.version = version;
      this.timeStamp = timeStamp;
      this.log(count + ' domains loaded');
    } catch (e) {
      this.log('Loading domains failed: ' + e);
    }
  },
  // --------------------------------------------------------------------------------------------------
  isDomain : function(domain, sub) {
    if (domain == sub) {
      return true;
    }
    var idx = sub.indexOf(domain);
    if (idx != -1) {
      var ddata = domain.split(/\./);
      var sdata = sub.split(/\./);
      var i = ddata.length - 1;
      var j = sdata.length - 1;
      var res = '';
      while (i >= 0 && j >= 0) {
        if (ddata[i] == sdata[j]) {
          if (res != '') {
            res = '.' + res;
          }
          res = ddata[i] + res;
        } else {
          break;
        }
        i--;
        j--;
      }
      if (res == domain) {
        return true;
      }
    }
    return false;
  },
  // --------------------------------------------------------------------------------------------------
  setOk : function(domain) {
    domain = domain.toLowerCase();
    for (var i = 0; i < this.okdomains.length; i++) {
      if (this.okdomains[i] == domain) {
        return;
      }
    }
    this.okdomains[this.okdomains.length] = domain;
  },
  // --------------------------------------------------------------------------------------------------
  isBlocked : function(domain) {
    domain = domain.toLowerCase();
    for (var i = 0; i < this.okdomains.length; i++) {
      if (this.okdomains[i] == domain) {
        return null;
      }
    }
    // matches of (sub)domains starting with the longest match
    var data = domain.split(/\./);
    for ( var i = 0; i < data.length - 1; i++) {
      var d = '';
      for ( var j = i; j < data.length; j++) {
        d += data[j];
        if (j < data.length - 1) {
          d += '.';
        }
      }
      for ( var k = 0; k < this.aadata.length; k++) {
        if (this.aadata[k].domain == d) {
          if (this.aadata[k].type == 'whitelist') {
            return null;
          } else {
            //dump('Blocked domain:' + k + ' ' + this.aadata[k].domain + '\n');
            return k;
          }
        }
      }
    }
    return null;
  },
  // --------------------------------------------------------------------------------------------------
  // This code was written by Tyler Akins and has been placed in the
  // public domain. It would be nice if you left this header intact.
  // Base64 code from Tyler Akins -- http://rumkin.com
  encode64 : function(input) {
    // hell of secure:
  input = '' + input;
  var keyStr = 'FEDCBAGHIJKLMNOPQRSTUVWXYZfedcbaghijklmnopqrstuvwxyz0123456789+/=';
  var output = '';
  var chr1, chr2, chr3;
  var enc1, enc2, enc3, enc4;
  var i = 0;

  do {
    chr1 = input.charCodeAt(i++);
    chr2 = input.charCodeAt(i++);
    chr3 = input.charCodeAt(i++);

    enc1 = chr1 >> 2;
    enc2 = ((chr1 & 3) << 4) | (chr2 >> 4);
    enc3 = ((chr2 & 15) << 2) | (chr3 >> 6);
    enc4 = chr3 & 63;

    if (isNaN(chr2)) {
      enc3 = enc4 = 64;
    } else if (isNaN(chr3)) {
      enc4 = 64;
    }

    output = output + keyStr.charAt(enc1) + keyStr.charAt(enc2)
        + keyStr.charAt(enc3) + keyStr.charAt(enc4);
  } while (i < input.length);

  return output;
},

decode64 : function(input) {
  input = '' + input;
  // hell of secure:
  var keyStr = 'FEDCBAGHIJKLMNOPQRSTUVWXYZfedcbaghijklmnopqrstuvwxyz0123456789+/=';
  var output = '';
  var chr1, chr2, chr3;
  var enc1, enc2, enc3, enc4;
  var i = 0;

  // remove all characters that are not A-Z, a-z, 0-9, +, /, or =
  input = input.replace(/[^A-Za-z0-9\+\/\=]/g, '');

  do {
    enc1 = keyStr.indexOf(input.charAt(i++));
    enc2 = keyStr.indexOf(input.charAt(i++));
    enc3 = keyStr.indexOf(input.charAt(i++));
    enc4 = keyStr.indexOf(input.charAt(i++));

    chr1 = (enc1 << 2) | (enc2 >> 4);
    chr2 = ((enc2 & 15) << 4) | (enc3 >> 2);
    chr3 = ((enc3 & 3) << 6) | enc4;

    output = output + String.fromCharCode(chr1);

    if (enc3 != 64) {
      output = output + String.fromCharCode(chr2);
    }
    if (enc4 != 64) {
      output = output + String.fromCharCode(chr3);
    }
  } while (i < input.length);

  return output;
}
};
